/* C++ code
  Security & Safety Alarm
  Simion Diana-Maria
*/

#include <Keypad.h>
#include <LiquidCrystal_I2C.h>
#include <Wire.h>

#define TRIG_PIN 12
#define ECHO_PIN 13
#define BUZZER_PIN 11
#define GAS_PIN A0
#define GREEN_LED_PIN A2
#define RED_LED_PIN A3

#define CM_DIST_CONSTANT 29.1
#define DISTANCE_SENSOR_LIMIT 50
#define GAS_SENSOR_LIMIT 300
#define ACTIVATE_ALARM_TIMEOUT 5
#define TRY_AGAIN_TIMEOUT 3

#define SERIAL_CURRENT_PASSWORD_MESSAGE "Your current password is: "
#define SERIAL_CORRECT_ENTERED_PASSWORD_MESSAGE                                \
  "The correct password was entered: "
#define SERIAL_WRONG_ENTERED_PASSWORD_MESSAGE                                  \
  "Some wrong password was entered: "
#define SERIAL_SMOKE_DETECTED_MESSAGE "Smoke detected! Be careful!"

#define NEXT_KEY 'A'
#define ACTIVATE_KEY 'B'
#define RESET_KEY 'C'

const byte ROWS = 4;
const byte COLS = 4;

char hexaKeys[ROWS][COLS] = {{'1', '2', '3', 'A'},
           {'4', '5', '6', 'B'},
           {'7', '8', '9', 'C'},
           {'*', '0', '#', 'D'}};

byte rowPins[ROWS] = {9, 8, 7, 6};
byte colPins[COLS] = {5, 4, 3, 2};

LiquidCrystal_I2C lcd(0x27, 16, 2);
long distInCm, duration;
float gasValue;
int roundedDist = 100;
Keypad customKeypad =
    Keypad(makeKeymap(hexaKeys), rowPins, colPins, ROWS, COLS);

int lcdcol = 0;
int lcdrow = 0;

char emptyLine[] = "                ";
char password[] = "AAAA";
char enteredPassword[] = "    ";
char setPasswordMessage[] = "Set password:";
char alarmOnMessage[] = "Password please:";
char activateMessage[] = "B -> Activate";
char resetMessage[] = "C -> Reset";
char pressAMessage[] = "Press A";
char text0[] = "Active in ";
char text1[] = "Try again ";
char *texts[2];

int passwordIndex = 0;
int MAX_INDEX = 4;

bool isPasswordSet = false;
bool isPasswordCorrect = false;
bool activeAlarm = false;
bool enteredPasswordState = false;
bool soundOn = false;

/* Initializes all the hardware components */
void setup()
{
  /* Serial setup */
  Serial.begin(9600);

  /* LCD setup */
  lcd.init();
  lcd.backlight();
  lcd.print("Starting");

  /* Distance sensor setup */
  pinMode(TRIG_PIN, OUTPUT);
  pinMode(ECHO_PIN, INPUT);

  /* Piezo buzzer setup */
  pinMode(BUZZER_PIN, OUTPUT);

  /* Leds setup */
  pinMode(GREEN_LED_PIN, OUTPUT);
  pinMode(RED_LED_PIN, OUTPUT);

  /* Keypad setup */
  customKeypad.setDebounceTime(50);

  /* Messages to be printed setup*/
  texts[0] = text0;
  texts[1] = text1;
}

/* Computes the rounded distance from sensor in centimeters */
void distanceSensorJob()
{
  /* The sensor is triggered by a HIGH pulse of 10 or more microseconds.
     Give a short LOW pulse beforehand to ensure a clean HIGH pulse: */
  digitalWrite(TRIG_PIN, LOW);
  delayMicroseconds(5);
  digitalWrite(TRIG_PIN, HIGH);
  delayMicroseconds(10);
  digitalWrite(TRIG_PIN, LOW);

  pinMode(ECHO_PIN, INPUT);
  duration = pulseIn(ECHO_PIN, HIGH);

  /* Convert the time into a distance */
  distInCm = (duration / 2) / CM_DIST_CONSTANT;
  roundedDist = floor(distInCm);
}

/* Checks gas value */
void gasSensorJob()
{
  gasValue = analogRead(GAS_PIN);

  if (gasValue > GAS_SENSOR_LIMIT) {
    tone(BUZZER_PIN, 523, 100);
  }
}

/* Triggers the buzzer based on the distance provided by the distance sensor */
void buzzerJob()
{
  if (roundedDist < DISTANCE_SENSOR_LIMIT) {
    soundOn = true;
  }

  if (activeAlarm && soundOn) {
    digitalWrite(BUZZER_PIN, HIGH);
    delay(80);
    digitalWrite(BUZZER_PIN, LOW);
    delay(80);
  }
}

/* Lights the green LED is the alarm is OFF and the red LED if the alarm in ON
 */
void ledsJob()
{
  if (activeAlarm == false) {
    digitalWrite(RED_LED_PIN, LOW);
    digitalWrite(GREEN_LED_PIN, HIGH);
  } else {
    digitalWrite(GREEN_LED_PIN, LOW);
    digitalWrite(RED_LED_PIN, HIGH);
  }
}

/* Prints helper messages on the LCD */
void lcdMessages()
{
  /* First row message */
  lcd.setCursor(0, 0);
  if (isPasswordSet == false) {
    lcd.print(setPasswordMessage);

    /* Second row message */
    lcd.setCursor(0, 1);
    lcd.print(pressAMessage);
  } else if (activeAlarm == true) {
    if (enteredPasswordState == false) {
      lcd.print(alarmOnMessage);
    }
  } else if (activeAlarm == false) {
    if (isPasswordCorrect == true) {
      lcd.print(activateMessage);

      /* Second row message */
      lcd.setCursor(0, 1);
      lcd.print(resetMessage);
    }
  }
}

/* Prints a message chosen with option parameter and a countdown in seconds
 * based on the timeout parameter */
void activateAlarmCountdown(int option, int timeout)
{
  int countdown = timeout;

  while (countdown != 0) {
    buzzerJob();
    buzzerJob();

    lcd.setCursor(0, 1);
    lcd.print(emptyLine);
    lcd.setCursor(0, 1);
    lcd.print(texts[option]);
    lcd.setCursor(10, 1);
    lcd.print(countdown);
    lcd.setCursor(13, 1);
    lcd.print("sec");

    countdown--;
    delay(1000);

    lcd.setCursor(10, 1);
    lcd.print("  ");
  }
}

/* Sets the current password */
void setPassword()
{
  while (true) {
    gasSensorJob();

    /* Helper messages for the user */
    if (passwordIndex < MAX_INDEX) {
      lcd.setCursor(lcdcol, 1);
      lcd.print("_");
    }
    if (passwordIndex == MAX_INDEX) {
      lcd.setCursor(6, 1);
      lcd.print("Press B");
    }

    char key = customKeypad.getKey();
    if (key) {
      if (passwordIndex < MAX_INDEX) {
        /* Show the current character and then replace
         * with a "*" */
        lcd.setCursor(lcdcol, 1);
        lcd.print(key);
        delay(150);
        lcd.setCursor(lcdcol, 1);
        lcd.print('*');

        /* Save the current character and advance */
        password[passwordIndex] = key;
        passwordIndex++;
        ++lcdcol;
      } else {
        if (key == ACTIVATE_KEY) {
          /* Send the complete password to the
           * user via bluetooth  */
          Serial.println(
              SERIAL_CURRENT_PASSWORD_MESSAGE);
          Serial.println(password);

          /* Prints a countdown until the alarm is
           * activated */
          activateAlarmCountdown(
              0, ACTIVATE_ALARM_TIMEOUT);

          /* Prints the messages for the next
           * state */
          lcd.setCursor(0, 0);
          lcd.print(alarmOnMessage);
          lcd.setCursor(0, 1);
          lcd.print(emptyLine);

          /* Mark the set alarm and reset indexes
           */
          isPasswordSet = true;
          activeAlarm = true;
          passwordIndex = 0;
          lcdcol = 0;

          /* The function can exit when the
           * activate key was pressed */
          return;
        }
      }
    }
  }
}

/* Checks if the entered password matches the current password and loops until
 * the correct password is entered */
void checkPassword()
{
  while (true) {
    /* Updte the buzzer and lcd status */
    buzzerJob();
    lcdMessages();
    gasSensorJob();

    /* Helper messages for the user */
    if (passwordIndex < MAX_INDEX) {
      lcd.setCursor(lcdcol, 1);
      lcd.print("_");
    }
    if (passwordIndex == MAX_INDEX) {
      lcd.setCursor(9, 1);
      lcd.print("Press A");
    }

    char key = customKeypad.getKey();
    if (key) {
      if (passwordIndex < MAX_INDEX) {
        /* Show the current character and then replace
         * with a "*" */
        lcd.setCursor(lcdcol, 1);
        lcd.print(key);
        delay(150);
        lcd.setCursor(lcdcol, 1);
        lcd.print('*');

        /* Save the current character and advance */
        enteredPassword[passwordIndex] = key;
        passwordIndex++;
        ++lcdcol;
      } else {
        /* The password was entered. The next state is
         * either succesful (the alarm deactivated),
         * either failed (try again state)
         */
        if (key == NEXT_KEY) {
          enteredPasswordState = true;

          if (strcmp(enteredPassword, password) ==
              0) {
            /* Send the correct password to
             * the user via bluetooth  */
            Serial.println(
                SERIAL_CORRECT_ENTERED_PASSWORD_MESSAGE);
            Serial.println(enteredPassword);

            /* Mark the correct password */
            activeAlarm = false;
            isPasswordCorrect = true;
            soundOn = false;
            lcd.clear();

            /* The function can exit when
             * the correct password was
             * entered */
            return;
          } else {
            Serial.println(
                SERIAL_WRONG_ENTERED_PASSWORD_MESSAGE);
            Serial.println(enteredPassword);

            /* Mark the incorrect password
             */
            enteredPasswordState = false;
            lcd.clear();
            passwordIndex = 0;
            lcdcol = 0;

            /* Prints the messages for the
             * next state */
            lcd.setCursor(1, 0);
            lcd.print(emptyLine);
            lcd.setCursor(4, 0);
            lcd.print("Wrong...");

            /* Prints a countdown until the
             * next try */
            activateAlarmCountdown(
                1, TRY_AGAIN_TIMEOUT);
            lcd.clear();
          }
        }
      }
    }
  }
}

/* Waits for an activate or reset action */
void activateOrReset()
{
  char key = customKeypad.getKey();

  if (key) {
    if (key == ACTIVATE_KEY) {
      /* Mark the activate state */
      activeAlarm = true;
      enteredPasswordState == false;
      isPasswordCorrect = false;
      passwordIndex = 0;
      lcdcol = 0;

      /* Prints the messages for the next state */
      lcd.setCursor(0, 0);
      lcd.print(alarmOnMessage);
      lcd.setCursor(0, 1);
      lcd.print(emptyLine);
    }

    if (key == RESET_KEY) {
      /* Mark the set alarm state */
      isPasswordSet = false;
      activeAlarm = false;
      enteredPasswordState == false;
      isPasswordCorrect = false;
      soundOn = false;
      passwordIndex = 0;
      lcdcol = 0;

      /* Prints the messages for the next state */
      lcd.setCursor(0, 0);
      lcd.print(setPasswordMessage);
      lcd.setCursor(0, 1);
      lcd.print(emptyLine);
    }
  }
}

void keypadAndLcdJob()
{
  /* Updte the buzzer and lcd status */
  buzzerJob();
  lcdMessages();
  gasSensorJob();

  /* Actions after the password was set */
  if (isPasswordSet) {
    if (activeAlarm) {
      checkPassword();
    } else {
      activateOrReset();
    }
  }

  /* The password has to be set */
  char key = customKeypad.getKey();
  if (key) {
    if (isPasswordSet == false) {
      if (key == NEXT_KEY) {
        lcd.setCursor(0, 1);
        lcd.print(emptyLine);
        setPassword();
      }
    }
  }
}

void loop()
{
  distanceSensorJob();

  gasSensorJob();

  ledsJob();

  keypadAndLcdJob();

  delay(100);
}
